/*
 * This file is part of piggy
 * Copyright (c) 2005 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
 *
 * This is the listener object it manages listening TCP sockets, for
 * each new connection that comes in we spawn off a new proxy object.
*/

#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "compiler.h"
#include "list.h"
#include "nbio.h"

struct listener {
	struct nbio io;
	listener_cbfn_t cbfn;
	void *priv;
};

static void listener_read(struct iothread *t, struct nbio *io)
{
	struct listener *l = (struct listener *)io;
	int fd;

	fd = accept(l->io.fd, NULL, 0);
	if ( fd < 0 ) {
		if ( errno == EAGAIN )
			nbio_inactive(t, &l->io);
		return;
	}

	if ( !socket_nonblock(fd) )
		return;

	l->cbfn(t, fd, l->priv);
}

static void listener_dtor(struct iothread *t, struct nbio *io)
{
	struct listener *l = (struct listener *)io;
	while ( close(l->io.fd) && (errno == EINTR) )
		/* do nothing */;
	free(l);
}

static struct nbio_ops listener_ops = {
	.read = listener_read,
	.dtor = listener_dtor,
};

struct nbio *do_listener(int domain, int proto,
				struct sockaddr *addr, socklen_t addrlen,
				listener_cbfn_t cb, void *priv)
{
	struct listener *l;

	l = calloc(1, sizeof(*l));
	if ( l == NULL )
		return NULL;

	INIT_LIST_HEAD(&l->io.list);

	l->cbfn = cb;
	l->priv = priv;

	l->io.fd = socket(domain, SOCK_STREAM, proto);
	if ( l->io.fd < 0 )
		goto err_free;

#if 1
	do{
		int val = 1;
		setsockopt(l->io.fd, SOL_SOCKET, SO_REUSEADDR,
				&val, sizeof(val));
	}while(0);
#endif

	if ( !socket_nonblock(l->io.fd) )
		goto err_close;

	if ( bind(l->io.fd, addr, addrlen) )
		goto err_close;

	if ( listen(l->io.fd, 64) )
		goto err_close;

	l->io.ops = &listener_ops;

	return &l->io;

err_close:
	while ( close(l->io.fd) && (errno == EINTR) )
		/* do nothing */;
err_free:
	free(l);
	return NULL;
}

struct nbio *listener_inet(int proto, uint32_t addr, uint16_t port,
				listener_cbfn_t cb, void *priv)
{
	struct sockaddr_in sa;
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(addr);
	sa.sin_port = htons(port);
	return do_listener(PF_INET, proto, (struct sockaddr *)&sa,
				sizeof(sa), cb, priv);
}

struct nbio *listener_unix(const char *fn,
				listener_cbfn_t cb, void *priv)
{
	struct sockaddr_un sa;
	unlink(fn);
	sa.sun_family = AF_UNIX;
	snprintf(sa.sun_path, sizeof(sa.sun_path), "%s", fn);
	return do_listener(PF_UNIX, 0, (struct sockaddr *)&sa,
				sizeof(sa), cb, priv);
}

void listener_add(struct iothread *t, struct nbio *io)
{
	nbio_add(t, io, NBIO_READ);
}
