/******************************************************************************/
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under gnu gpl v2                                                  */
/* tissue api                                                                 */
/******************************************************************************/

#include "tissue.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <arpa/inet.h>
#include <netinet/sctp.h>
#include <netdb.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <err.h>

#define BACKLOG 100

static Tissue *tcopy = NULL;  /* copy of tissue for sigint handler*/

struct tsclient {
	struct nbio io;
	struct list_head list;
	Tissue *tissue;
};

static void sigint_handler(int s)
{
	free_tissue(tcopy);
	exit(EXIT_SUCCESS);
}

static int tcmp(const void *A, const void *B)
{
	struct itimer *a = *(struct itimer **)A;
	struct itimer *b = *(struct itimer **)B;
	if ( a->next_time < b->next_time )
		return -1;
	if ( a->next_time > b->next_time )
		return 1;
	return 0;
}

static void update_timers(Tissue *t)
{
	if ( t->max_timers < t->num_timers ) {
		struct itimer *it;
		unsigned int i = 0;
		void *new;

		new = realloc(t->prio, t->num_timers * sizeof(*t->prio));
		if ( new == NULL )
			logfatal_tissue("couldn't allocate memory for timers");
		t->prio = new;
		t->max_timers = t->num_timers;

		list_for_each_entry(it, &t->timers, list) {
			t->prio[i] = it;
			i++;
		}
	}

	qsort(t->prio, t->num_timers, sizeof(*t->prio), tcmp);
}

void add_timer(Tissue *t, uint64_t interval, timer_cbfn_t cb)
{
	struct itimer *it;

	it = calloc(1, sizeof(*it));
	if ( it == NULL )
		logfatal_tissue("couldn't allocate memory for timer");

	it->interval = interval;
	it->next_time = interval;
	it->cb = cb;
	list_add_tail(&it->list, &t->timers);
	t->num_timers++;
	update_timers(t);
}

static void do_timers(Tissue *t, uint64_t ts)
{
	unsigned int i;

	if ( ts < t->prev_time )
		return;

	t->prev_time = ts;

	if ( t->num_timers == 0 )
		return;
	
	for(i = 0; i < t->num_timers; i++) {
		struct itimer *it = t->prio[i];
		unsigned int n;

		if ( ts < it->next_time )
			break;

		//printf("Fired timer %u %llu > %llu ", i,
		//	ts, it->next_time);

		if ( it->next_time == it->interval ) {
			it->next_time = ts + it->interval;
			n = 1;
		}else{
			n = (unsigned int )((ts - it->next_time) /
				it->interval);
			n++;
			it->next_time += n * it->interval;
		}

		//printf("%u times\n", n);

		do {
			it->cb(t);
		}while(--n);
	}

	if ( i )
		update_timers(t);
}

static int client_ds(Tissue *t, char *buf, size_t len)
{
	unsigned int i, j;
	DS *ds;

	if(!(ds = pkt2ds(buf, len)))
		return 0;

	if ( t->time_sync )
		do_timers(t, tv2ts(&ds->tm));

	/* set tissue from datastream */
	/* set cytokine levels */
	for(i = 0; i < ds->num_signals; i++) {
		if( (ds->signal[i]->id >= 1) &&
			(ds->signal[i]->id < (int) t->num_cytokines) )
			switch(t->cytokine_type[ds->signal[i]->id]) {
			/* replace current value in t with incoming */
			case CYTOKINE_REFLECTIVE:
				t->cytokine[ds->signal[i]->id]->val = \
						ds->signal[i]->val;
				break;

			case CYTOKINE_ADDITIVE:
			/* add incoming to current value in t */
				t->cytokine[ds->signal[i]->id]->val += \
						ds->signal[i]->val;
				break;

			/* new is a function of current and incoming */
			case CYTOKINE_CUSTOM:
				break;

			default:
				logtime_tissue("unknown cytokine type");
				break;
			}
	}

	/* set antigen in tissue */
	/* create multiple copies according to antigen_multiplier */
	for(j = 0; j < t->antigen_multiplier; j++)
		for(i = 0; i < ds->num_antigen; i++) {
			unsigned int ind;

			/* random insertion */
			ind = rand() % t->num_antigen;
			if(t->antigen[ind])
				free_antigen(t->antigen[ind]);
			t->antigen[ind] = dup_antigen(ds->antigen[i]);
		}

	if(t->log)
		log_ds(ds);

	/* tidy up */
	free_ds(ds);
	return 1;
}

static void do_time_sync(Tissue *tissue)
{
	struct itimer *it;

	nbio_del(&tissue->nbio, tissue->remote_svr);
	nbio_del(&tissue->nbio, tissue->local_svr);
	tissue->time_sync = 1;
	tissue->prev_time = 0;

	list_for_each_entry(it, &tissue->timers, list)
		it->next_time = it->interval;
}

/* handle tissue client once connected */
static void client_read(struct iothread *t, struct nbio *io)
{
	struct tsclient *c = (struct tsclient *)io;
	static char buf[TS_BUFFER_SIZE];
	Tissue *tissue = c->tissue;
	static int type;
	int len;

	/* receive data */
again:
	len = recv(c->io.fd, buf, TS_BUFFER_SIZE, MSG_NOSIGNAL);
	if(len <= 0) {
		if ( len < 0 && errno == EAGAIN ) {
			nbio_inactive(&tissue->nbio, &c->io);
			return;
		}
		goto err;
	}

	/* process data */
	type = buf[0];
	switch(type) {
	case CLIENT_DS:
		/* 0x01 - receive datastream event */
		if ( !client_ds(tissue, buf, len) )
			goto err;
		break;
	case CLIENT_MSG:
		logtime_tissue(buf + 1);
		break;
	case CLIENT_TIME_SYNC:
		if ( tissue->num_clients != 1 )
			goto err;
		do_time_sync(tissue);
		//printf("client %i is time syncer\n", c->io.fd);
		break;
	default:
		goto err;
	}
	goto again;
err:
	nbio_del(&tissue->nbio, &c->io);
	return;
}

static void client_dtor(struct iothread *t, struct nbio *io)
{
	struct tsclient *c = (struct tsclient *)io;

	c->tissue->num_clients--;
	//printf("client %i closed num_clients=%u\n",
	//	c->io.fd, c->tissue->num_clients);
	list_del(&c->list);
	close(c->io.fd);
	free(c);
}

static const struct nbio_ops client_ops = {
	.read = client_read,
	.dtor = client_dtor,
};

static void accept_cb(struct iothread *t, int s, void *priv)
{
	struct tsclient *c;
	Tissue *tissue = priv;

	c = calloc(1, sizeof(*c));
	if ( c == NULL )
		close(s);

	c->io.fd = s;
	c->io.ops = &client_ops,
	c->tissue = tissue;
	list_add_tail(&c->list, &tissue->clients);
	tissue->num_clients++;

	//printf("Accepted client connection fd=%i num_clients=%u\n",
	//	s, tissue->num_clients);

	nbio_add(t, &c->io, NBIO_READ);
}

/* initialise the tissue */
Tissue *init_tissue(const uint16_t server_port,
		const char *sockname,
		const unsigned int num_antigen,
		const unsigned int num_cytokines,
		const unsigned int num_cells)
{
	Tissue *tissue;
	static int done;
	unsigned int i;
	struct sigaction siga;
	struct nbio *l;

	srand(time(NULL));
	srand48(time(NULL));
	
	if ( done == 0 ) {
		eventloop_poll_ctor();
		eventloop_epoll_ctor();
		done = 1;
	}

	/* setup logfile */
	open_logfile();
	logtime_tissue("starting tissue server");

	/* allocate memory */
	if(!(tissue = calloc(1, sizeof(Tissue))))
		logfatal_tissue("couldn't allocate memory for tissue");

	if ( !nbio_init(&tissue->nbio, NULL) )
		logfatal_tissue("couldn't initialize nonblocking I/O");

	tissue->num_antigen = num_antigen;
	if(tissue->num_antigen)
		if(!(tissue->antigen = calloc(tissue->num_antigen, sizeof(Antigen *))))
			logfatal_tissue("couldn't allocate memory for antigen");
	tissue->num_cytokines = num_cytokines;
	if(tissue->num_cytokines) {
		if(!(tissue->cytokine = calloc(tissue->num_cytokines, \
				sizeof(Cytokine *))))
			logfatal_tissue("couldn't allocate memory for cytokines");
		if(!(tissue->cytokine_type = calloc(tissue->num_cytokines, \
				sizeof(int))))
			logfatal_tissue("couldn't allocate memory for cytokine types");
		for(i = 0; i < tissue->num_cytokines; i++) {
			tissue->cytokine[i] = init_cytokine();
			tissue->cytokine[i]->id = i;
			tissue->cytokine_type[i] = DEF_CYTOKINE_TYPE;
		}
	}
	tissue->num_cells = num_cells;
	if(tissue->num_cells)
		if(!(tissue->cell = calloc(tissue->num_cells, sizeof(Cell *))))
			logfatal_tissue("couldn't allocate memory for cells");

	/* capture INT to exit cleanly */
	tcopy = tissue;
	siga.sa_handler = sigint_handler;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = SA_RESTART;
	if(sigaction(SIGINT, &siga, NULL) == -1)
		logfatal_tissue("couldn't setup sigint handler");

	/* set defaults */
	tissue->log = 0;  /* no logging of server */
	tissue->cytokine_decay_rate = DEF_DECAY_RATE;  /* cytokine decay rate */
	tissue->antigen_multiplier = 1;  /* how many copies of incoming antigen */

	l = listener_inet(IPPROTO_SCTP, INADDR_ANY,
				(server_port) ? server_port : TS_DEFAULT_PORT,
				accept_cb, tissue);
	if ( l == NULL )
		logfatal_tissue("couldn't create remote server socket");
	tissue->remote_svr = l;
	listener_add(&tissue->nbio, l);

	if ( sockname ) {
		l = listener_unix(sockname, accept_cb, tissue);
		if ( l == NULL )
			logfatal_tissue("couldn't create local server socket");

		tissue->local_svr = l;
		listener_add(&tissue->nbio, l);
	}

	INIT_LIST_HEAD(&tissue->clients);
	INIT_LIST_HEAD(&tissue->timers);
	return tissue;
}

static unsigned int do_tmo_calc(Tissue *t)
{
	uint64_t tmo;

	if ( t->num_timers == 0 )
		return -1;

	if ( t->prio[0]->next_time < t->prev_time )
		return 0;

	tmo = t->prio[0]->next_time - t->prev_time;
	return tmo / 1000;
}

static unsigned int calc_tmo(Tissue *t)
{
	if ( t->time_sync ) {
		return -1;
	}else{
		struct timeval tv;
		gettimeofday(&tv, NULL);
		do_timers(t, tv2ts(&tv));
		return do_tmo_calc(t);
	}
}

static void set_initial_time(Tissue *t)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	t->prev_time = tv2ts(&tv);
}

void ts_mainloop(Tissue *t)
{
	set_initial_time(t);

	t->time_sync = 0;
	do {
		unsigned int tmo;
		tmo = calc_tmo(t);
		nbio_pump(&t->nbio, tmo);
	}while( !list_empty(&t->nbio.active) ||
		!list_empty(&t->nbio.inactive) );
	t->time_sync = 0;
}

/* free memory for tissue structure */
void free_tissue(Tissue *tissue)
{
	logtime_tissue("stopping tissue server");

	if(!tissue)
		return;

	/* close server */
	nbio_fini(&tissue->nbio);

	/* close log */
	close_logfile();

	/* TODO: clean up memory from timers and clients... */
	free(tissue);
}

void print_tissue(Tissue *tissue)
{
	unsigned int i;

	printf("tissue_cytokines=%d: ", tissue->num_cytokines);
	for(i = 0; i < tissue->num_cytokines; i++)
		printf(" (%d,%f)", tissue->cytokine[i]->id, tissue->cytokine[i]->val);
	printf("\ntissue_antigen=%d: ", tissue->num_antigen);
	for(i = 0; i < tissue->num_antigen; i++)
		if(tissue->antigen[i]) {
			printf(" (%d,", tissue->antigen[i]->id);
			fwrite((void *) tissue->antigen[i]->val, sizeof(char), \
					tissue->antigen[i]->len, stdout);
			printf(")");
		}
	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[i]) {
			printf("\ntissue_cell_%d: ", i);
			print_cell(tissue->cell[i]);
		}
}

void print_tissue_cytokines(Tissue *tissue)
{
	unsigned int i;

	printf("tissue_cytokines:");
	for(i = 0; i < tissue->num_cytokines; i++)
		printf(" (%d,%f)", tissue->cytokine[i]->id, tissue->cytokine[i]->val);
	printf("\n");
}

void print_tissue_antigen(Tissue *tissue)
{
	unsigned int i, count = 0;

	printf("tissue_antigen:");
	for(i = 0; i < tissue->num_antigen; i++)
		if(tissue->antigen[i]) {
			printf(" (%d,", tissue->antigen[i]->id);
			fwrite((void *) tissue->antigen[i]->val, sizeof(char), \
					tissue->antigen[i]->len, stdout);
			printf(")");
			count++;
		}
	printf(" total_antigen=%d\n", count);
}

void print_tissue_cells(Tissue *tissue)
{
	unsigned int i, count = 0;

	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[i]) {
			printf("tissue_cell_%d: ", i);
			print_cell(tissue->cell[i]);
			count++;
		}
	printf("total_tissue_cells=%d\n", count);
}

/* set the input of a cell from the tissue */
void set_cell_input(Tissue *tissue, const unsigned int index)
{
	Cell *cell  = tissue->cell[index];
	unsigned int i, j, ind1, ind2;

	/* handle antigen receptors: transfer antigen from tissue to cell */
	for(i = 0; i < cell->num_antigen_receptors; i++) {
		/* get random antigen in tissue */
		ind1 = (int) (((double) tissue->num_antigen * rand()) / \
				(RAND_MAX + 1.0));
		/* only copy if antigen available */
		if(tissue->antigen[ind1]) {
			/* randomly copy to cell antigen store */
			ind2 = (int) (((double) cell->num_antigen * rand()) / \
					(RAND_MAX + 1.0));
			/* remove any previous antigen */
			free_antigen(cell->antigen[ind2]);
			/* copy antigen from cell to tissue */
			cell->antigen[ind2] = dup_antigen(tissue->antigen[ind1]);
			/* remove antigen from tissue */
			free_antigen(tissue->antigen[ind1]);
			tissue->antigen[ind1] = NULL;
		}
	}

	/* handle cytokine receptors: sample cytokine levels */
	for(i = 0; i < cell->num_cytokine_receptors; i++)
		/* only sample if cytokine available in tissue */
		if((cell->cytokine_receptor[i]->id < (int) tissue->num_cytokines))
			cell->cytokine_receptor[i]->val = add_noise( \
					tissue->cytokine[cell->cytokine_receptor[i]->id]->val, \
					cell->cytokine_receptor[i]->noise);

	/* handle cell receptors: look for other cells */
	for(i = 0; i < cell->num_cell_receptors; i++) {
		cell->cell_receptor[i]->cell = NULL;
		if(drand48() > cell->cell_receptor[i]->efficiency)
			continue;
		/* get random cell in tissue */
		ind1 = (int) (((double) (tissue->num_cells - 1) * rand()) / \
				(RAND_MAX + 1.0));
		if(ind1 >= index)  /* don't attach receptor to self */
			ind1++;
		if(!tissue->cell[ind1])
			continue;
		/* only conjunct with cell if same type and state as receptor likes */
		if((cell->cell_receptor[i]->type == -1) \
				|| cell->cell_receptor[i]->type == tissue->cell[ind1]->type)
			if((cell->cell_receptor[i]->state == -1) \
					|| cell->cell_receptor[i]->state == \
					tissue->cell[ind1]->state)
				cell->cell_receptor[i]->cell = tissue->cell[ind1];
	}

	/* handle vr receptors: match antigen */
	for(i = 0; i < cell->num_vr_receptors; i++) {
		/* can only examine antigen from cell conjoined via cell receptor */
		cell->vr_receptor[i]->cell = NULL;
		cell->vr_receptor[i]->activated = 0;
		free_antigen(cell->vr_receptor[i]->key);
		cell->vr_receptor[i]->key = NULL;
		randomise_list_2(cell->num_cell_receptors);
		for(j = 0; j < cell->num_cell_receptors; j++) {
			if(!cell->cell_receptor[randlist2[j]]->cell)
				continue;
			if((cell->vr_receptor[i]->type != -1) && \
					(cell->vr_receptor[i]->type != \
					cell->cell_receptor[randlist2[j]]->cell->type))
				continue;
			if((cell->cell_receptor[randlist2[j]]->state != -1) && \
					(cell->vr_receptor[i]->state != \
					cell->cell_receptor[randlist2[j]]->cell->state))
				continue;
			cell->vr_receptor[i]->cell = \
					cell->cell_receptor[randlist2[j]]->cell;
			break;  /* matching cell found */
		}
		if(!cell->vr_receptor[i]->cell)
			continue;  /* no matching cell found for receptor */
		if(!cell->vr_receptor[i]->cell->num_antigen_producers) {
			/* matching cell isn't producing antigen */
			cell->vr_receptor[i]->cell = NULL;
			continue;
		}
		/* see if antigen matches any of those on antigen producers */
		randomise_list_2(cell->vr_receptor[i]->cell->num_antigen_producers);
		for(j = 0; j < cell->vr_receptor[i]->cell->num_antigen_producers; j++)
			if(cell->vr_receptor[i]->cell->antigen_producer[randlist2[j]]->antigen)
				if(cell->vr_receptor[i]->match(cell->vr_receptor[i]->lock, \
						cell->vr_receptor[i]->cell->antigen_producer[randlist2[j]]->antigen) >= cell->vr_receptor[i]->threshold) {
					cell->vr_receptor[i]->activated = 1;
					cell->vr_receptor[i]->key = dup_antigen( \
					  cell->vr_receptor[i]->cell->antigen_producer[randlist2[j]]->antigen);
					break;
				}
	}

	/* reset response producers */
	for(i = 0; i < cell->num_response_producers; i++) {
		free_antigen(cell->response_producer[i]->antigen);
		cell->response_producer[i]->antigen = NULL;
	}
}

/* set the tissue from output of a cell */
void set_tissue(Tissue *tissue, const unsigned int index)
{
	Cell *cell  = tissue->cell[index];
	unsigned int i;

	/* adjust cytokine levels from cell cytokine producers */
	for(i = 0; i < cell->num_cytokine_producers; i++)
		if(cell->cytokine_producer[i]->type < (int) tissue->num_cytokines)
			tissue->cytokine[cell->cytokine_producer[i]->type]->val += \
					cell->cytokine_producer[i]->val;
}

/* allow tissue and cells to interact */
void step_tissue(Tissue *tissue)
{
	unsigned int i;

	/* set input to cell */
	randomise_list(tissue->num_cells);
	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[randlist[i]]) {
			tissue->cell[randlist[i]]->iterations++;
			set_cell_input(tissue, randlist[i]);
			set_antigen_producers(tissue->cell[randlist[i]]);
		}

	/* cycle cells */
	randomise_list(tissue->num_cells);
	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[randlist[i]])
			tissue->cell[randlist[i]]->cycle(tissue->cell[randlist[i]]);

	/* set tissue from cell output */
	randomise_list(tissue->num_cells);
	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[randlist[i]]) {
			set_tissue(tissue, randlist[i]);
			produce_response(tissue->cell[randlist[i]]);
		}

	/* reduce cytokine levels in tissue */
	for(i = 0; i < tissue->num_cytokines; i++)
		tissue->cytokine[i]->val *= tissue->cytokine_decay_rate;
}

/* try and set a value from the parameter file */
int read_parameter_uint(char *filename, char *parameter, unsigned int *val)
{
	FILE *fp;
	char line[1024], *ptr, *endptr[1];

	if(!(fp = fopen(filename, "r")))
		err(1, "error: couldn't read parameter file");

	while(fgets(line, sizeof(line), fp)) {
		if(line[0] == '#')  /* comment line */
			continue;
		if(!(ptr = index(line, '=')))
			err(1, "error: couldn't parse parameter file");
		*ptr = '\0';
		if(strcmp(line, parameter) == 0) {
			/* it's our parameter */
			*val = (unsigned int) strtol(ptr + 1, endptr, 10);
			if(**endptr != '\n')
				err(1, "edddrror: couldn't parse parameter file");
			fclose(fp);
			return 0;  /* found */
		}
	}

	fclose(fp);
	return 1;  /* not found */
}

/* save parameters in parameter file to log */
void log_parameter_file(char *filename)
{
	FILE *fp;
	char line[128], message[1024], *ptr;

	if(!(fp = fopen(filename, "r")))
		snprintf(message, sizeof(message) - 1, "using default parameters");
	else {
		ptr = message;
		ptr += snprintf(message, sizeof(message) - 1, \
				"using parameter file %s: ", filename);
		while(fgets(line, sizeof(line), fp)) {
			if(line[0] == '#')  /* comment line */
				continue;
			line[strlen(line) - 1] = '\0';
			ptr += sprintf(ptr, "%s ; ", line);  /* fix me */
		}
		*(ptr - 2) = '\0';
		fclose(fp);
	}

	/* save parameters to log */
	logtime_tissue(message);
}
